import pytest
import os
import json

from psc_auth_integration.entities.auth_user_session import AuthUserSession
from assignments.common.utils import setup_data, cleanup, get_policies_summary, DEFENSE_URL, USERNAME, PASSWORD, ENV, \
    LOCAL_PAS_HOST, EXPECTED_CALL_TO_SUCCEED, ASSET_GROUPS, ENDPOINTS

SESSION = None
POLICIES = None

'''
Policies:
    - Monitored
    - Standard
    - Advanced
    - e2e-test-policy-1
    - e2e-test-policy-2
    - e2e-test-policy-3
    - e2e-test-policy-4 
    - e2e-test-policy-5
    
Asset Groups:
    - ag-criteria-endpoint-1_2_3:
        policy: Monitored (Rank 1)
        query: name: "do-not-update-linux-1" OR name:"do-not-update-linux-2" OR name:"do-not-update-linux-3"
    - ag-criteria-linux:
        policy: Standard (Rank 2)
        query: os.equals: "LINUX"
    - ag-criteria-windows:
        policy: Advanced (Rank 3)
        query: os.equals: "WINDOWS"
    - ag-policy-1:
        policy: e2e-test-policy-1 (Rank 4)
    - ag-policy-2:
        policy: e2e-test-policy-2 (Rank 5)
    - ag-policy-3:
        policy: e2e-test-policy-3 (Rank 6)
    - ag-policy-4:
        policy: e2e-test-policy-4 (Rank 7)
    - ag-policy-5:
        policy: e2e-test-policy-5 (Rank 8)
    - ag-policy-monitored:
        policy: Monitored (Rank 1)

Endpoints:
    - do-not-update-1:
        asset_groups:
            - ag-criteria-windows
            - ag-policy-1
    - do-not-update-2:
        asset_groups:
            - ag-criteria-windows
            - ag-policy-2
    - do-not-update-3:
        asset_groups:
            - ag-criteria-windows
            - ag-policy-3
    - do-not-update-4:
        asset_groups:
            - ag-policy-monitored
            - ag-criteria-windows
            - ag-policy-3
    - do-not-update-5:
        asset_groups:
            - ag-criteria-windows
            - ag-policy-5
    - do-not-update-6:
        asset_groups:
            - ag-criteria-windows
            - ag-policy-1
            - ag-policy-2
    - do-not-update-7:
        asset_groups:
            - ag-criteria-windows
            - ag-policy-3
            - ag-policy-4
    - do-not-update-8:
        override: Monitored (Rank 1)
        asset_groups:
            - ag-criteria-windows
            - ag-policy-1
            - ag-policy-2
            - ag-policy-3
            - ag-policy-4
            - ag-policy-5
    - do-not-update-linux-1
        asset_groups:
            - ag-criteria-endpoint-1_2_3
            - ag-criteria-linux
    - do-not-update-linux-2
        asset_groups:
            - ag-criteria-endpoint-1_2_3
            - ag-criteria-linux
    - do-not-update-linux-3
        override: Advanced
        asset_groups:
            - ag-criteria-endpoint-1_2_3
            - ag-criteria-linux
    - do-not-update-linux-4
        asset_groups:
            - ag-criteria-endpoint-1_2_3
    - do-not-update-linux-5
        asset_groups:
            - ag-criteria-endpoint-1_2_3
    - do-not-update-linux-6
        asset_groups:
            - ag-criteria-endpoint-1_2_3
    - do-not-update-linux-7
        asset_groups:
            - ag-criteria-endpoint-1_2_3
    - do-not-update-linux-8
        asset_groups:
            - ag-criteria-endpoint-1_2_3
    - pas-e2e-default-endpoint
        os: MAC
'''


@pytest.fixture(scope="module", autouse=True)
def init():
    global SESSION, POLICIES
    os.environ["ENV"] = ENV
    SESSION = AuthUserSession(gateway_url=DEFENSE_URL, username=USERNAME, password=PASSWORD)
    setup_data(SESSION)
    POLICIES = get_policies_summary(SESSION).get("policies")


@pytest.mark.preview
def test_policy_policies_change_preview_rank_change():
    payload = {
        "policies": [
            {
                "id": POLICIES[3].get('id'),
                "position": 3
            },
        ]
    }
    previews = preview_policies_change(SESSION, payload)

    assert 1 == len(previews)

    assert 2 == previews[0]["asset_count"]
    assert 3 == previews[0]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[0]["current_policy"]["id"]
    assert 3 == previews[0]["new_policy"]["position"]
    assert POLICIES[3].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[0]["asset_query"])
    assert 2 == len(affected_asset_ids)
    assert set([19037736, 19037730]) == set(affected_asset_ids)


@pytest.mark.preview
def test_policy_policies_change_preview_asset_group_policy_rank_change_update_to_lower_rank():
    payload = {
        "policies": [
            {
                "id": POLICIES[0].get('id'),
                "position": POLICIES[-1].get('position')
            },
        ]
    }

    previews = preview_policies_change(SESSION, payload)

    assert 2 == len(previews)

    assert 2 == previews[0]["asset_count"]
    assert 1 == previews[0]["current_policy"]["position"]
    assert POLICIES[0].get('id') == previews[0]["current_policy"]["id"]
    assert 1 == previews[0]["new_policy"]["position"]
    assert POLICIES[1].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[0]["asset_query"])
    assert 2 == len(affected_asset_ids)
    assert set([19037753, 19037752]) == set(affected_asset_ids)

    assert 1 == previews[1]["asset_count"]
    assert 1 == previews[1]["current_policy"]["position"]
    assert POLICIES[0].get('id') == previews[1]["current_policy"]["id"]
    assert 2 == previews[1]["new_policy"]["position"]
    assert POLICIES[2].get('id') == previews[1]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[1]["asset_query"])
    assert 1 == len(affected_asset_ids)
    assert set([19037734]) == set(affected_asset_ids)


@pytest.mark.preview
def test_policy_policies_change_preview_rank_change_unchanged_effective_policy():
    payload = {
        "policies": [
            {
                "id": POLICIES[2].get('id'),
                "position": 2
            },
        ]
    }
    previews = preview_policies_change(SESSION, payload)

    assert 0 == len(previews)


@pytest.mark.preview
def test_preview_asset_groups_change_add_policy_override():
    do_not_update_7_endpoint = None

    for endpoint in ENDPOINTS:
        if endpoint["name"] == "do-not-update-7":
            do_not_update_7_endpoint = endpoint
            break

    assert do_not_update_7_endpoint is not None, "endpoint do-not-update-7 not found."

    payload = {
        "action": "ADD_POLICY_OVERRIDE",
        "asset_ids": [do_not_update_7_endpoint["id"]],
        "policy_id": POLICIES[0].get('id')
    }
    previews = preview_asset_group_change(SESSION, payload)

    assert 1 == len(previews)
    assert 1 == previews[0]["asset_count"]
    assert 3 == previews[0]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[0]["current_policy"]["id"]
    assert 1 == previews[0]["new_policy"]["position"]
    assert POLICIES[0].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[0]["asset_query"])
    assert 1 == len(affected_asset_ids)
    assert set([19037737]) == set(affected_asset_ids)


@pytest.mark.preview
def test_preview_asset_groups_change_add_policy_override_with_search_definition():
    payload = {
        "action": "ADD_POLICY_OVERRIDE",
        "assets_search_definition": {
            "criteria": {
                "os": ["WINDOWS"]
            }
        },
        "policy_id": POLICIES[7].get('id')
    }
    previews = preview_asset_group_change(SESSION, payload)

    assert 2 == len(previews)

    assert 2 == previews[0]["asset_count"]
    assert 1 == previews[0]["current_policy"]["position"]
    assert POLICIES[0].get('id') == previews[0]["current_policy"]["id"]
    assert 8 == previews[0]["new_policy"]["position"]
    assert POLICIES[7].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, search_definition=previews[0]["assets_search_definition"])
    assert 2 == len(affected_asset_ids)
    assert set([19037739, 19037734]) == set(affected_asset_ids)

    assert 6 == previews[1]["asset_count"]
    assert 3 == previews[1]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[1]["current_policy"]["id"]
    assert 8 == previews[1]["new_policy"]["position"]
    assert POLICIES[7].get('id') == previews[1]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, search_definition=previews[1]["assets_search_definition"])
    assert 6 == len(affected_asset_ids)
    assert set([19037737, 19037736, 19037735, 19037732, 19037731, 19037730]) == set(affected_asset_ids)

    payload = {
        "action": "ADD_POLICY_OVERRIDE",
        "assets_search_definition": {
            "criteria": {
                "os": ["WINDOWS", "LINUX"]
            }
        },
        "policy_id": POLICIES[7].get('id')
    }
    previews = preview_asset_group_change(SESSION, payload)

    assert 3 == len(previews)

    assert 4 == previews[0]["asset_count"]
    assert 1 == previews[0]["current_policy"]["position"]
    assert POLICIES[0].get('id') == previews[0]["current_policy"]["id"]
    assert 8 == previews[0]["new_policy"]["position"]
    assert POLICIES[7].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, search_definition=previews[0]["assets_search_definition"])
    assert 4 == len(affected_asset_ids)
    assert set([19037753, 19037752, 19037739, 19037734]) == set(affected_asset_ids)

    assert 5 == previews[1]["asset_count"]
    assert 2 == previews[1]["current_policy"]["position"]
    assert POLICIES[1].get('id') == previews[1]["current_policy"]["id"]
    assert 8 == previews[1]["new_policy"]["position"]
    assert POLICIES[7].get('id') == previews[1]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, search_definition=previews[1]["assets_search_definition"])
    assert 5 == len(affected_asset_ids)
    assert set([19037761, 19037760, 19037758, 19037757, 19037756]) == set(affected_asset_ids)

    assert 7 == previews[2]["asset_count"]
    assert 3 == previews[2]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[2]["current_policy"]["id"]
    assert 8 == previews[2]["new_policy"]["position"]
    assert POLICIES[7].get('id') == previews[2]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, search_definition=previews[2]["assets_search_definition"])
    assert 7 == len(affected_asset_ids)
    assert set([19037755, 19037737, 19037736, 19037735, 19037732, 19037731, 19037730]) == set(affected_asset_ids)

    payload = {
        "action": "ADD_POLICY_OVERRIDE",
        "assets_search_definition": {
            "query": "do-not-update-linux"
        },
        "policy_id": POLICIES[7].get('id')
    }
    previews = preview_asset_group_change(SESSION, payload)

    assert 3 == len(previews)

    assert 2 == previews[0]["asset_count"]
    assert 1 == previews[0]["current_policy"]["position"]
    assert POLICIES[0].get('id') == previews[0]["current_policy"]["id"]
    assert 8 == previews[0]["new_policy"]["position"]
    assert POLICIES[7].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, search_definition=previews[0]["assets_search_definition"])
    assert 2 == len(affected_asset_ids)
    assert set([19037753, 19037752]) == set(affected_asset_ids)

    assert 5 == previews[1]["asset_count"]
    assert 2 == previews[1]["current_policy"]["position"]
    assert POLICIES[1].get('id') == previews[1]["current_policy"]["id"]
    assert 8 == previews[1]["new_policy"]["position"]
    assert POLICIES[7].get('id') == previews[1]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, search_definition=previews[1]["assets_search_definition"])
    assert 5 == len(affected_asset_ids)
    assert set([19037761, 19037760, 19037758, 19037757, 19037756]) == set(affected_asset_ids)

    assert 1 == previews[2]["asset_count"]
    assert 3 == previews[2]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[2]["current_policy"]["id"]
    assert 8 == previews[2]["new_policy"]["position"]
    assert POLICIES[7].get('id') == previews[2]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, search_definition=previews[2]["assets_search_definition"])
    assert 1 == len(affected_asset_ids)
    assert set([19037755]) == set(affected_asset_ids)


@pytest.mark.preview
def test_preview_asset_groups_change_add_policy_override_to_asset_having_policy_override():
    do_not_update_8_endpoint = None

    for endpoint in ENDPOINTS:
        if endpoint["name"] == "do-not-update-8":
            do_not_update_8_endpoint = endpoint
            break

    assert do_not_update_8_endpoint is not None, "endpoint do-not-update-8 not found."

    payload = {
        "action": "ADD_POLICY_OVERRIDE",
        "asset_ids": [do_not_update_8_endpoint["id"]],
        "policy_id": POLICIES[5].get('id')
    }
    previews = preview_asset_group_change(SESSION, payload)

    assert 1 == len(previews)
    assert 1 == previews[0]["asset_count"]
    assert 1 == previews[0]["current_policy"]["position"]
    assert POLICIES[0].get('id') == previews[0]["current_policy"]["id"]
    assert 6 == previews[0]["new_policy"]["position"]
    assert POLICIES[5].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[0]["asset_query"])
    assert 1 == len(affected_asset_ids)
    assert set([19037739]) == set(affected_asset_ids)


@pytest.mark.preview
def test_preview_asset_groups_change_remove_policy_override():
    do_not_update_8_endpoint = None

    for endpoint in ENDPOINTS:
        if endpoint["name"] == "do-not-update-8":
            do_not_update_8_endpoint = endpoint
            break

    assert do_not_update_8_endpoint is not None, "endpoint do-not-update-8 not found."

    payload = {
        "action": "REMOVE_POLICY_OVERRIDE",
        "asset_ids": [do_not_update_8_endpoint["id"]]
    }
    previews = preview_asset_group_change(SESSION, payload)

    assert 1 == len(previews)
    assert 1 == previews[0]["asset_count"]
    assert 1 == previews[0]["current_policy"]["position"]
    assert POLICIES[0].get('id') == previews[0]["current_policy"]["id"]
    assert 3 == previews[0]["new_policy"]["position"]
    assert POLICIES[2].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[0]["asset_query"])
    assert 1 == len(affected_asset_ids)
    assert set([19037739]) == set(affected_asset_ids)


@pytest.mark.preview
def test_preview_asset_groups_change_manual_membership_add_members():
    do_not_update_6_endpoint = None
    for endpoint in ENDPOINTS:
        if endpoint["name"] == "do-not-update-6":
            do_not_update_6_endpoint = endpoint
            break

    assert do_not_update_6_endpoint is not None, "endpoint do-not-update-6 not found."

    asset_group_policy_monitored = None
    for group in ASSET_GROUPS:
        if group["name"] == "ag-policy-monitored":
            asset_group_policy_monitored = group
            break

    assert asset_group_policy_monitored is not None, "endpoint ag-policy-monitored not found."

    payload = {
        "action": "ADD_MEMBERS",
        "asset_ids": [do_not_update_6_endpoint["id"]],
        "asset_group_ids": [asset_group_policy_monitored["id"]]
    }
    previews = preview_asset_group_change(SESSION, payload)

    assert 1 == len(previews)
    assert 1 == previews[0]["asset_count"]
    assert 3 == previews[0]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[0]["current_policy"]["id"]
    assert 1 == previews[0]["new_policy"]["position"]
    assert POLICIES[0].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[0]["asset_query"])
    assert 1 == len(affected_asset_ids)
    assert set([19037736]) == set(affected_asset_ids)


@pytest.mark.preview
def test_preview_asset_groups_change_manual_membership_remove_members():
    do_not_update_4_endpoint = None
    for endpoint in ENDPOINTS:
        if endpoint["name"] == "do-not-update-4":
            do_not_update_4_endpoint = endpoint
            break

    assert do_not_update_4_endpoint is not None, "endpoint do-not-update-4 not found."

    asset_group_policy_monitored = None
    for group in ASSET_GROUPS:
        if group["name"] == "ag-policy-monitored":
            asset_group_policy_monitored = group
            break

    assert asset_group_policy_monitored is not None, "endpoint ag-policy-monitored not found."

    payload = {
        "action": "REMOVE_MEMBERS",
        "asset_ids": [do_not_update_4_endpoint["id"]],
        "asset_group_ids": [asset_group_policy_monitored["id"]]
    }
    previews = preview_asset_group_change(SESSION, payload)

    assert 1 == len(previews)
    assert 1 == previews[0]["asset_count"]
    assert 1 == previews[0]["current_policy"]["position"]
    assert POLICIES[0].get('id') == previews[0]["current_policy"]["id"]
    assert 3 == previews[0]["new_policy"]["position"]
    assert POLICIES[2].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[0]["asset_query"])
    assert 1 == len(affected_asset_ids)
    assert set([19037734]) == set(affected_asset_ids)


@pytest.mark.preview
def test_preview_asset_groups_change_asset_group_policy_change():
    asset_group_criteria_windows = None
    for group in ASSET_GROUPS:
        if group["name"] == "ag-criteria-windows":
            asset_group_criteria_windows = group
            break

    assert asset_group_criteria_windows is not None, "endpoint ag-criteria-windows not found."

    payload = {
        "action": "ASSET_GROUPS_UPDATE",
        "asset_group_ids": [asset_group_criteria_windows["id"]],
        "asset_group_query": asset_group_criteria_windows["query"],
        "policy_id": POLICIES[4].get('id')
    }
    previews = preview_asset_group_change(SESSION, payload)

    assert 2 == len(previews)

    assert 4 == previews[0]["asset_count"]
    assert 3 == previews[0]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[0]["current_policy"]["id"]
    assert 5 == previews[0]["new_policy"]["position"]
    assert POLICIES[4].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[0]["asset_query"])
    assert 4 == len(affected_asset_ids)
    assert set([19037737, 19037735, 19037732, 19037731]) == set(affected_asset_ids)

    assert 2 == previews[1]["asset_count"]
    assert 3 == previews[1]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[1]["current_policy"]["id"]
    assert 4 == previews[1]["new_policy"]["position"]
    assert POLICIES[3].get('id') == previews[1]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[1]["asset_query"])
    assert 2 == len(affected_asset_ids)
    assert set([19037736, 19037730]) == set(affected_asset_ids)


@pytest.mark.preview
def test_preview_asset_groups_change_asset_group_dynamic_criteria_change():
    asset_group_criteria_windows = None
    for group in ASSET_GROUPS:
        if group["name"] == "ag-criteria-windows":
            asset_group_criteria_windows = group
            break

    assert asset_group_criteria_windows is not None, "endpoint ag-criteria-windows not found."

    payload = {
        "action": "ASSET_GROUPS_UPDATE",
        "asset_group_ids": [asset_group_criteria_windows["id"]],
        "asset_group_query": "os:LINUX",
        "policy_id": POLICIES[2].get('id')
    }
    previews = preview_asset_group_change(SESSION, payload)

    assert 4 == len(previews)

    assert 1 == previews[0]["asset_count"]
    assert 3 == previews[0]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[0]["current_policy"]["id"]
    assert 8 == previews[0]["new_policy"]["position"]
    assert POLICIES[7].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[0]["asset_query"])
    assert 1 == len(affected_asset_ids)
    assert set([19037735]) == set(affected_asset_ids)

    assert 1 == previews[1]["asset_count"]
    assert 3 == previews[1]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[1]["current_policy"]["id"]
    assert 5 == previews[1]["new_policy"]["position"]
    assert POLICIES[4].get('id') == previews[1]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[1]["asset_query"])
    assert 1 == len(affected_asset_ids)
    assert set([19037731]) == set(affected_asset_ids)

    assert 2 == previews[2]["asset_count"]
    assert 3 == previews[2]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[2]["current_policy"]["id"]
    assert 4 == previews[2]["new_policy"]["position"]
    assert POLICIES[3].get('id') == previews[2]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[2]["asset_query"])
    assert 2 == len(affected_asset_ids)
    assert set([19037736, 19037730]) == set(affected_asset_ids)

    assert 2 == previews[3]["asset_count"]
    assert 3 == previews[3]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[3]["current_policy"]["id"]
    assert 6 == previews[3]["new_policy"]["position"]
    assert POLICIES[5].get('id') == previews[3]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[3]["asset_query"])
    assert 2 == len(affected_asset_ids)
    assert set([19037737, 19037732]) == set(affected_asset_ids)


@pytest.mark.preview
def test_preview_asset_groups_change_asset_group_remove_policy():
    asset_group_criteria_windows = None
    for group in ASSET_GROUPS:
        if group["name"] == "ag-criteria-windows":
            asset_group_criteria_windows = group
            break

    assert asset_group_criteria_windows is not None, "endpoint ag-criteria-windows not found."

    payload = {
        "action": "ASSET_GROUPS_UPDATE",
        "asset_group_ids": [asset_group_criteria_windows["id"]],
        "asset_group_query": asset_group_criteria_windows["query"],
    }
    previews = preview_asset_group_change(SESSION, payload)

    assert 4 == len(previews)

    assert 1 == previews[0]["asset_count"]
    assert 3 == previews[0]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[0]["current_policy"]["id"]
    assert 8 == previews[0]["new_policy"]["position"]
    assert POLICIES[7].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[0]["asset_query"])
    assert 1 == len(affected_asset_ids)
    assert set([19037735]) == set(affected_asset_ids)

    assert 1 == previews[1]["asset_count"]
    assert 3 == previews[1]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[1]["current_policy"]["id"]
    assert 5 == previews[1]["new_policy"]["position"]
    assert POLICIES[4].get('id') == previews[1]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[1]["asset_query"])
    assert 1 == len(affected_asset_ids)
    assert set([19037731]) == set(affected_asset_ids)

    assert 2 == previews[2]["asset_count"]
    assert 3 == previews[2]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[2]["current_policy"]["id"]
    assert 4 == previews[2]["new_policy"]["position"]
    assert POLICIES[3].get('id') == previews[2]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[2]["asset_query"])
    assert 2 == len(affected_asset_ids)
    assert set([19037736, 19037730]) == set(affected_asset_ids)

    assert 2 == previews[3]["asset_count"]
    assert 3 == previews[3]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[3]["current_policy"]["id"]
    assert 6 == previews[3]["new_policy"]["position"]
    assert POLICIES[5].get('id') == previews[3]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[3]["asset_query"])
    assert 2 == len(affected_asset_ids)
    assert set([19037737, 19037732]) == set(affected_asset_ids)


@pytest.mark.preview
def test_preview_asset_groups_change_asset_group_policy_with_lower_rank():
    asset_group_criteria_windows = None
    for group in ASSET_GROUPS:
        if group["name"] == "ag-criteria-windows":
            asset_group_criteria_windows = group
            break

    assert asset_group_criteria_windows is not None, "endpoint ag-criteria-windows not found."

    payload = {
        "action": "ASSET_GROUPS_UPDATE",
        "asset_group_ids": [asset_group_criteria_windows["id"]],
        "asset_group_query": asset_group_criteria_windows["query"],
        "policy_id": POLICIES[len(POLICIES) - 1].get('id')
    }
    previews = preview_asset_group_change(SESSION, payload)

    assert 4 == len(previews)

    assert 1 == previews[0]["asset_count"]
    assert 3 == previews[0]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[0]["current_policy"]["id"]
    assert 8 == previews[0]["new_policy"]["position"]
    assert POLICIES[7].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[0]["asset_query"])
    assert 1 == len(affected_asset_ids)
    assert set([19037735]) == set(affected_asset_ids)

    assert 1 == previews[1]["asset_count"]
    assert 3 == previews[1]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[1]["current_policy"]["id"]
    assert 5 == previews[1]["new_policy"]["position"]
    assert POLICIES[4].get('id') == previews[1]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[1]["asset_query"])
    assert 1 == len(affected_asset_ids)
    assert set([19037731]) == set(affected_asset_ids)

    assert 2 == previews[2]["asset_count"]
    assert 3 == previews[2]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[2]["current_policy"]["id"]
    assert 4 == previews[2]["new_policy"]["position"]
    assert POLICIES[3].get('id') == previews[2]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[2]["asset_query"])
    assert 2 == len(affected_asset_ids)
    assert set([19037736, 19037730]) == set(affected_asset_ids)

    assert 2 == previews[3]["asset_count"]
    assert 3 == previews[3]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[3]["current_policy"]["id"]
    assert 6 == previews[3]["new_policy"]["position"]
    assert POLICIES[5].get('id') == previews[3]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[3]["asset_query"])
    assert 2 == len(affected_asset_ids)
    assert set([19037737, 19037732]) == set(affected_asset_ids)


@pytest.mark.preview
def test_preview_asset_groups_change_asset_group_policy_with_lower_rank_should_not_affect_associations():
    asset_group_policy_2 = None
    for group in ASSET_GROUPS:
        if group["name"] == "ag-policy-2":
            asset_group_policy_2 = group
            break

    assert asset_group_policy_2 is not None, "endpoint ag-policy-2 not found."

    payload = {
        "action": "ASSET_GROUPS_UPDATE",
        "asset_group_ids": [asset_group_policy_2["id"]],
        "policy_id": POLICIES[len(POLICIES) - 1].get('id')
    }
    previews = preview_asset_group_change(SESSION, payload)

    assert 0 == len(previews)


@pytest.mark.preview
def test_preview_asset_groups_change_asset_group_policy_with_lower_rank_add_new_asset_via_dynamic_query():
    asset_group_policy_monitored = None
    for group in ASSET_GROUPS:
        if group["name"] == "ag-policy-monitored":
            asset_group_policy_monitored = group
            break

    assert asset_group_policy_monitored is not None, "endpoint ag-policy-monitored not found."

    payload = {
        "action": "ASSET_GROUPS_UPDATE",
        "asset_group_ids": [asset_group_policy_monitored["id"]],
        "asset_group_query": "name:do-not-update-3",
        "policy_id": asset_group_policy_monitored["policy_id"]
    }
    previews = preview_asset_group_change(SESSION, payload)

    assert 1 == len(previews)

    assert 1 == previews[0]["asset_count"]
    assert 3 == previews[0]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[0]["current_policy"]["id"]
    assert 1 == previews[0]["new_policy"]["position"]
    assert POLICIES[0].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, previews[0]["asset_query"])
    assert 1 == len(affected_asset_ids)
    assert set([19037732]) == set(affected_asset_ids)


@pytest.mark.preview
def test_preview_asset_groups_change_asset_group_remove_dynamic_criteria():
    asset_group_criteria_monitored_linux_endpoints = None
    for group in ASSET_GROUPS:
        if group["name"] == "ag-criteria-endpoint-1_2_3":
            asset_group_criteria_monitored_linux_endpoints = group
            break

    assert asset_group_criteria_monitored_linux_endpoints is not None, "endpoint ag-criteria-endpoint-1_2_3 not found."

    payload = {
        "action": "ASSET_GROUPS_UPDATE",
        "asset_group_ids": [asset_group_criteria_monitored_linux_endpoints["id"]],
        "policy_id": asset_group_criteria_monitored_linux_endpoints["policy_id"]
    }
    previews = preview_asset_group_change(SESSION, payload)

    assert 1 == len(previews)

    assert 2 == previews[0]["asset_count"]
    assert 1 == previews[0]["current_policy"]["position"]
    assert POLICIES[0].get('id') == previews[0]["current_policy"]["id"]
    assert 2 == previews[0]["new_policy"]["position"]
    assert POLICIES[1].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[0]["asset_query"])
    assert 2 == len(affected_asset_ids)
    assert set([19037753, 19037752]) == set(affected_asset_ids)


@pytest.mark.preview
def test_preview_asset_groups_change_asset_group_add_asset_with_higher_ranked_policy():
    asset_group_policy_5 = None
    for group in ASSET_GROUPS:
        if group["name"] == "ag-policy-5":
            asset_group_policy_5 = group
            break

    assert asset_group_policy_5 is not None, "endpoint ag-policy-5 not found."

    payload = {
        "action": "ASSET_GROUPS_UPDATE",
        "asset_group_ids": [asset_group_policy_5["id"]],
        "policy_id": asset_group_policy_5["policy_id"],
        "asset_group_query": "name:pas-e2e-default-endpoint"
    }
    previews = preview_asset_group_change(SESSION, payload)

    assert 1 == len(previews)

    assert 1 == previews[0]["asset_count"]
    assert 2 == previews[0]["current_policy"]["position"]
    assert POLICIES[1].get('id') == previews[0]["current_policy"]["id"]
    assert 8 == previews[0]["new_policy"]["position"]
    assert POLICIES[7].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[0]["asset_query"])
    assert 1 == len(affected_asset_ids)
    assert set([19245811]) == set(affected_asset_ids)


@pytest.mark.preview
def test_preview_asset_groups_change_asset_group_remove_policy_and_dynamic_criteria():
    asset_group_criteria_windows = None
    for group in ASSET_GROUPS:
        if group["name"] == "ag-criteria-windows":
            asset_group_criteria_windows = group
            break

    assert asset_group_criteria_windows is not None, "endpoint ag-criteria-windows not found."

    payload = {
        "action": "ASSET_GROUPS_UPDATE",
        "asset_group_ids": [asset_group_criteria_windows["id"]]
    }
    previews = preview_asset_group_change(SESSION, payload)

    assert 4 == len(previews)

    assert 1 == previews[0]["asset_count"]
    assert 3 == previews[0]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[0]["current_policy"]["id"]
    assert 8 == previews[0]["new_policy"]["position"]
    assert POLICIES[7].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[0]["asset_query"])
    assert 1 == len(affected_asset_ids)
    assert set([19037735]) == set(affected_asset_ids)

    assert 1 == previews[1]["asset_count"]
    assert 3 == previews[1]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[1]["current_policy"]["id"]
    assert 5 == previews[1]["new_policy"]["position"]
    assert POLICIES[4].get('id') == previews[1]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[1]["asset_query"])
    assert 1 == len(affected_asset_ids)
    assert set([19037731]) == set(affected_asset_ids)

    assert 2 == previews[2]["asset_count"]
    assert 3 == previews[2]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[2]["current_policy"]["id"]
    assert 4 == previews[2]["new_policy"]["position"]
    assert POLICIES[3].get('id') == previews[2]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[2]["asset_query"])
    assert 2 == len(affected_asset_ids)
    assert set([19037736, 19037730]) == set(affected_asset_ids)

    assert 2 == previews[3]["asset_count"]
    assert 3 == previews[3]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[3]["current_policy"]["id"]
    assert 6 == previews[3]["new_policy"]["position"]
    assert POLICIES[5].get('id') == previews[3]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[3]["asset_query"])
    assert 2 == len(affected_asset_ids)
    assert set([19037737, 19037732]) == set(affected_asset_ids)


@pytest.mark.preview
def test_preview_asset_groups_delete():
    asset_group_criteria_windows = None
    for group in ASSET_GROUPS:
        if group["name"] == "ag-criteria-windows":
            asset_group_criteria_windows = group
            break

    assert asset_group_criteria_windows is not None, "endpoint ag-criteria-windows not found."

    payload = {
        "action": "ASSET_GROUPS_DELETE",
        "asset_group_ids": [asset_group_criteria_windows["id"]]
    }
    previews = preview_asset_group_change(SESSION, payload)

    assert 4 == len(previews)

    assert 1 == previews[0]["asset_count"]
    assert 3 == previews[0]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[0]["current_policy"]["id"]
    assert 8 == previews[0]["new_policy"]["position"]
    assert POLICIES[7].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[0]["asset_query"])
    assert 1 == len(affected_asset_ids)
    assert set([19037735]) == set(affected_asset_ids)

    assert 1 == previews[1]["asset_count"]
    assert 3 == previews[1]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[1]["current_policy"]["id"]
    assert 5 == previews[1]["new_policy"]["position"]
    assert POLICIES[4].get('id') == previews[1]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[1]["asset_query"])
    assert 1 == len(affected_asset_ids)
    assert set([19037731]) == set(affected_asset_ids)

    assert 2 == previews[2]["asset_count"]
    assert 3 == previews[2]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[2]["current_policy"]["id"]
    assert 4 == previews[2]["new_policy"]["position"]
    assert POLICIES[3].get('id') == previews[2]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[2]["asset_query"])
    assert 2 == len(affected_asset_ids)
    assert set([19037736, 19037730]) == set(affected_asset_ids)

    assert 2 == previews[3]["asset_count"]
    assert 3 == previews[3]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[3]["current_policy"]["id"]
    assert 6 == previews[3]["new_policy"]["position"]
    assert POLICIES[5].get('id') == previews[3]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[3]["asset_query"])
    assert 2 == len(affected_asset_ids)
    assert set([19037737, 19037732]) == set(affected_asset_ids)


@pytest.mark.preview
def test_preview_asset_groups_change_update_policy_and_criteria():
    asset_group_criteria_windows = None
    for group in ASSET_GROUPS:
        if group["name"] == "ag-criteria-windows":
            asset_group_criteria_windows = group
            break

    assert asset_group_criteria_windows is not None, "endpoint ag-criteria-windows not found."

    payload = {
        "action": "ASSET_GROUPS_UPDATE",
        "asset_group_ids": [asset_group_criteria_windows["id"]],
        "asset_group_query": "os:LINUX",
        "policy_id": POLICIES[0].get('id')
    }
    previews = preview_asset_group_change(SESSION, payload)

    assert 5 == len(previews)

    assert 5 == previews[0]["asset_count"]
    assert 2 == previews[0]["current_policy"]["position"]
    assert POLICIES[1].get('id') == previews[0]["current_policy"]["id"]
    assert 1 == previews[0]["new_policy"]["position"]
    assert POLICIES[0].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[0]["asset_query"])
    assert 5 == len(affected_asset_ids)
    assert set([19037761, 19037760, 19037758, 19037757, 19037756]) == set(affected_asset_ids)

    assert 1 == previews[1]["asset_count"]
    assert 3 == previews[1]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[1]["current_policy"]["id"]
    assert 8 == previews[1]["new_policy"]["position"]
    assert POLICIES[7].get('id') == previews[1]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[1]["asset_query"])
    assert 1 == len(affected_asset_ids)
    assert set([19037735]) == set(affected_asset_ids)

    assert 1 == previews[2]["asset_count"]
    assert 3 == previews[2]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[2]["current_policy"]["id"]
    assert 5 == previews[2]["new_policy"]["position"]
    assert POLICIES[4].get('id') == previews[2]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[2]["asset_query"])
    assert 1 == len(affected_asset_ids)
    assert set([19037731]) == set(affected_asset_ids)

    assert 2 == previews[3]["asset_count"]
    assert 3 == previews[3]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[3]["current_policy"]["id"]
    assert 4 == previews[3]["new_policy"]["position"]
    assert POLICIES[3].get('id') == previews[3]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[3]["asset_query"])
    assert 2 == len(affected_asset_ids)
    assert set([19037736, 19037730]) == set(affected_asset_ids)

    assert 2 == previews[4]["asset_count"]
    assert 3 == previews[4]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[4]["current_policy"]["id"]
    assert 6 == previews[4]["new_policy"]["position"]
    assert POLICIES[5].get('id') == previews[4]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[4]["asset_query"])
    assert 2 == len(affected_asset_ids)
    assert set([19037737, 19037732]) == set(affected_asset_ids)


@pytest.mark.preview
def test_preview_asset_groups_create():
    payload = {
        "action": "ASSET_GROUPS_CREATE",
        "asset_group_query": "os:WINDOWS",
        "policy_id": POLICIES[0].get('id')
    }
    previews = preview_asset_group_change(SESSION, payload)

    assert 1 == len(previews)
    assert 6 == previews[0]["asset_count"]
    assert 3 == previews[0]["current_policy"]["position"]
    assert POLICIES[2].get('id') == previews[0]["current_policy"]["id"]
    assert 1 == previews[0]["new_policy"]["position"]
    assert POLICIES[0].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[0]["asset_query"])
    assert 6 == len(affected_asset_ids)
    assert set([19037737, 19037736, 19037735, 19037732, 19037731, 19037730]) == set(affected_asset_ids)


@pytest.mark.preview
def test_preview_asset_groups_create_should_not_affect_policy_overrides():
    payload = {
        "action": "ASSET_GROUPS_CREATE",
        "asset_group_query": 'name: "do-not-update-linux-3" OR name:"do-not-update-linux-4" OR name:"do-not-update-linux-5"',
        "policy_id": POLICIES[0].get('id')
    }
    previews = preview_asset_group_change(SESSION, payload)

    assert 1 == len(previews)
    assert 2 == previews[0]["asset_count"]
    assert 2 == previews[0]["current_policy"]["position"]
    assert POLICIES[1].get('id') == previews[0]["current_policy"]["id"]
    assert 1 == previews[0]["new_policy"]["position"]
    assert POLICIES[0].get('id') == previews[0]["new_policy"]["id"]

    affected_asset_ids = search_assets(SESSION, query=previews[0]["asset_query"])
    assert 2 == len(affected_asset_ids)
    assert set([19037757, 19037756]) == set(affected_asset_ids)


def preview_policies_change(session, payload):
    response = session.post(
        url=f"{LOCAL_PAS_HOST}/v1/orgs/{session.get_user_org_key()}/policies/preview",
        data=json.dumps(payload)
    )
    assert response.ok, EXPECTED_CALL_TO_SUCCEED
    return json.loads(response.content).get("preview")


def preview_asset_group_change(session, payload):
    response = session.post(
        url=f"{LOCAL_PAS_HOST}/v1/orgs/{session.get_user_org_key()}/asset-groups/preview",
        data=json.dumps(payload)
    )
    assert response.ok, EXPECTED_CALL_TO_SUCCEED
    return json.loads(response.content).get("preview")


def search_assets(session, query=None, search_definition=None):
    if query:
        payload = {
            "query": query
        }
    else:
        payload = search_definition

    response = session.post(
        url=f"{session.gateway_url}/asset-registry/v1/appservices/v6/orgs/{session.get_user_org_key()}/devices/_search",
        data=json.dumps(payload)
    )
    assert response.ok, EXPECTED_CALL_TO_SUCCEED
    results = json.loads(response.content).get("results")

    return [ result.get("id") for result in results ]
